"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ConsentPlugin = void 0;
var _ = require("..");
var _SegmentDestination = require("./SegmentDestination");
const CONSENT_PREF_UPDATE_EVENT = 'Segment Consent Preference';
/**
 * This plugin interfaces with the consent provider and it:
 *
 * - stamps all events with the consent metadata.
 * - augments all destinations with a consent filter plugin that prevents events from reaching them if
 * they are not compliant current consent setup
 * - listens for consent change from the provider and notifies Segment
 */
class ConsentPlugin extends _.Plugin {
  type = _.PluginType.before;
  constructor(consentCategoryProvider, categories) {
    super();
    this.consentCategoryProvider = consentCategoryProvider;
    this.categories = categories;
  }
  configure(analytics) {
    super.configure(analytics);
    analytics.getPlugins().forEach(this.injectConsentFilterIfApplicable);
    analytics.onPluginLoaded(this.injectConsentFilterIfApplicable);
    this.consentCategoryProvider.setApplicableCategories(this.categories);
    this.consentCategoryProvider.onConsentChange(categoryPreferences => {
      this.analytics?.track(CONSENT_PREF_UPDATE_EVENT, {
        consent: {
          categoryPreferences
        }
      }).catch(e => {
        throw e;
      });
    });
  }
  async execute(event) {
    if (this.isConsentUpdateEvent(event)) {
      return event;
    }
    event.context = {
      ...event.context,
      consent: {
        categoryPreferences: await this.consentCategoryProvider.getConsentStatus()
      }
    };
    return event;
  }
  shutdown() {
    this.consentCategoryProvider.shutdown?.();
  }
  injectConsentFilterIfApplicable = plugin => {
    if (this.isDestinationPlugin(plugin)) {
      plugin.add(new ConsentFilterPlugin(event => {
        const allCategories = this.analytics?.consentSettings.get()?.allCategories || [];
        const settings = this.analytics?.settings.get() || {};
        const preferences = event.context?.consent?.categoryPreferences || {};
        if (plugin.key === _SegmentDestination.SEGMENT_DESTINATION_KEY) {
          const noneConsented = allCategories.every(category => !preferences[category]);
          return this.isConsentUpdateEvent(event) || !this.isConsentFeatureSetup() || !(noneConsented && !this.hasUnmappedDestinations());
        }
        const integrationSettings = settings?.[plugin.key];
        if (this.containsConsentSettings(integrationSettings)) {
          const categories = integrationSettings.consentSettings.categories;
          return !this.isConsentUpdateEvent(event) && categories.every(category => preferences?.[category]);
        }
        return true;
      }));
    }
  };
  isDestinationPlugin(plugin) {
    return plugin.type === _.PluginType.destination;
  }
  containsConsentSettings = settings => {
    return typeof settings?.consentSettings?.categories === 'object';
  };
  isConsentUpdateEvent(event) {
    return event.event === CONSENT_PREF_UPDATE_EVENT;
  }
  hasUnmappedDestinations() {
    return this.analytics?.consentSettings.get()?.hasUnmappedDestinations === true;
  }
  isConsentFeatureSetup() {
    return typeof this.analytics?.consentSettings.get() === 'object';
  }
}

/**
 * This plugin reads the consent metadata set on the context object and then drops the events
 * if they are going into a destination which violates's set consent preferences
 */
exports.ConsentPlugin = ConsentPlugin;
class ConsentFilterPlugin extends _.Plugin {
  type = _.PluginType.before;
  constructor(shouldAllowEvent) {
    super();
    this.shouldAllowEvent = shouldAllowEvent;
  }
  execute(event) {
    return this.shouldAllowEvent(event) ? event : undefined;
  }
}
//# sourceMappingURL=ConsentPlugin.js.map